from pathlib import Path
from typing import List, Optional
from collections import deque
import pandas as pd
import sys
import logging
from .utils.rvd import get_RVD_seq
from .utils.sequence import get_sequence

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s - %(message)s",
    stream=sys.stdout,
)
logger = logging.getLogger("talenWF")

COMPLEMENT = {"T": "A", "A": "T", "C": "G", "G": "C"}


def findAll(seq, sub, start=0, end=None):
    if end is None:
        end = len(seq)
    return [i for i in range(start, end) if seq[i : i + len(sub)] == sub]


class FindTALTask:
    """TALEN window finder class."""

    def __init__(
        self,
        fasta: Optional[str] = None,
        sequence: Optional[str] = None,
        min_spacer: int = 14,
        max_spacer: int = 18,
        array_min: int = 14,
        array_max: int = 18,
        outpath: str = "talenWF_out",
        filter_base: Optional[int] = None,
        upstream_bases: List[str] = ["T"],
        gspec: bool = False,
    ):
        """
        Initialize the TAL window finder.

        Args:
            fasta: The FASTA file to process.
            sequence: The sequence to process.
            min_spacer: The minimum spacer length.
            max_spacer: The maximum spacer length.
            array_min: The minimum array length.
            array_max: The maximum array length.
            filter_base: The base position to filter. If None, no filtering is done. (1-indexed)
            upstream_bases: The upstream bases to consider. (Default: ['T'])
            outpath: The path to the output file.
            gspec: Whether to use G-specific RVDs. (Default: False)
        """
        self.fasta = fasta
        self.sequence = sequence
        self.min_spacer = min_spacer
        self.max_spacer = max_spacer
        self.array_min = array_min
        self.array_max = array_max
        self.outpath = outpath
        self.filter_base = (
            filter_base - 1 if filter_base is not None else None
        )  # make it 0-based
        self.upstream_bases = upstream_bases
        self.gspec = gspec
        # Calculate distances
        self.max_dist = (
            2 * array_max + max_spacer
        )  # max dist between TAL1 and TAL2 start/end
        self.min_dist = (
            2 * array_min + min_spacer
        )  # min dist between TAL1 and TAL2 start/end

    def _find_tal_pairs_for_filterpos(self):
        for upstream_base in self.upstream_bases:
            downstream_base = COMPLEMENT[upstream_base]
            up_pos_array = findAll(
                self.sequence,
                upstream_base,
                max(0, self.filter_base - (self.max_dist // 2) - 1),
                self.filter_base - (self.min_dist // 2) + 1,
            )
            down_pos_array = findAll(
                self.sequence,
                downstream_base,
                self.filter_base + ((self.min_dist) // 2),
                min(
                    len(self.sequence),
                    self.filter_base + ((self.max_dist + 1) // 2) + 2,
                ),
            )
            for up_pos in up_pos_array:
                for down_pos in down_pos_array:
                    assert self.sequence[up_pos] == upstream_base
                    assert self.sequence[down_pos] == downstream_base
                    window_len = (down_pos - 1) - (up_pos + 1) + 1
                    if window_len < self.min_dist or window_len > self.max_dist:
                        continue
                    for spacer_length in range(self.min_spacer, self.max_spacer + 1):
                        # if TAL array lengths are within range
                        sum_tal_lengths = window_len - spacer_length
                        if (
                            sum_tal_lengths >= 2 * self.array_min
                            and sum_tal_lengths <= 2 * self.array_max
                        ):
                            yield self._create_tal_pair(
                                self.filter_base, up_pos, down_pos, spacer_length
                            )

    def _find_tal_pairs_for_seq(self):
        """Generator function that yields TAL pairs for a sequence."""

        if self.filter_base is not None:
            yield from self._find_tal_pairs_for_filterpos()

        # No filter base -> find all valid TAL pairs within the sequence
        # Storing upstream base occurences in a queue to visit the upmost first
        else:
            for upstream_base in self.upstream_bases:
                upstream_q = deque()
                for j in range(len(self.sequence)):
                    ch = self.sequence[j]
                    if ch == upstream_base:
                        upstream_q.append(j)
                    elif ch == COMPLEMENT[upstream_base]:
                        # remove any upstream candidates that are too far for j (i < j-max_dist)
                        up_pos_min = j - self.max_dist - 1
                        up_pos_max = j - self.min_dist - 1
                        while upstream_q:
                            up_pos = upstream_q[0]
                            if up_pos > up_pos_max:
                                break
                            upstream_q.popleft()
                            if up_pos < up_pos_min:
                                continue
                            # up_pos is in range up_pos_min to up_pos_max
                            for spacer_length in range(
                                self.min_spacer, self.max_spacer + 1
                            ):
                                sum_tal_lengths = j - up_pos - 1 - spacer_length
                                for tal1_length in range(
                                    self.array_min, self.array_max + 1
                                ):
                                    tal2_length = sum_tal_lengths - tal1_length
                                    if (
                                        tal2_length >= self.array_min
                                        and tal2_length <= self.array_max
                                    ):
                                        spacer_start = up_pos + 1 + tal1_length
                                        cut = spacer_start + spacer_length // 2
                                        yield self._create_tal_pair(
                                            cut, up_pos, j, spacer_length, False
                                        )

    def _create_tal_pair(
        self,
        cut: int,
        up_pos: int,
        down_pos: int,
        spacer_length: int,
        use_middle2: bool = True,
    ):
        """Create a TAL pair attributes dictionary from positions for the output table."""
        rows = []
        tal1_start = up_pos + 1
        if (spacer_length % 2 == 1) or not use_middle2:
            tal1_ends = [cut - spacer_length // 2 - 1]
        else:
            # if even-sized spacer cut site in the middle 2 positions
            tal1_ends = [cut - spacer_length // 2 - 1, cut - spacer_length // 2]
        for tal1_end in tal1_ends:
            tal2_start = tal1_end + spacer_length + 1
            tal2_end = down_pos - 1

            tal1 = self.sequence[tal1_start : tal1_end + 1]
            tal2 = self.sequence[tal2_start : tal2_end + 1]
            spacer = self.sequence[tal1_end + 1 : tal2_start]
            if (
                len(tal1) > self.array_max
                or len(tal2) > self.array_max
                or len(tal1) < self.array_min
                or len(tal2) < self.array_min
            ):
                continue
            # Plus strand sequence format example: T AGAGCT(TAL1) cgcagcgt(spacer) ACGTCGAC(TAL2) A
            plus_seq = f"{self.sequence[up_pos]} {tal1} {spacer.lower()} {tal2} {self.sequence[down_pos]}"
            row = {
                "Sequence Name": self.sequence_id,
                "Cut Site": cut,
                "TAL1 start": tal1_start,
                "TAL2 start": tal2_end,
                "TAL1 length": len(tal1),
                "TAL2 length": len(tal2),
                "Spacer length": spacer_length,
                "Spacer range": f"{tal1_end+1}-{tal2_start-1}",
                "TAL1 RVDs": get_RVD_seq(tal1, self.gspec),
                "TAL2 RVDs": get_RVD_seq(tal2, self.gspec),
                "Plus strand sequence": plus_seq,
                "Unique RE sites in spacer": "",  # TODO: add function for RE sites in spacer
                "% RVDs HD or NN/NH": "",  # TODO: add function for % RVDs HD or NN/NH
            }
            rows.append(row)
        return rows

    def run(self) -> Optional[pd.DataFrame]:
        """
        Runs the TAL window finder.

        Returns:
            A DataFrame containing the TAL windows.
        """
        self.sequence, self.sequence_id = get_sequence(self.fasta, self.sequence)
        tal_window_rows: List[dict] = []
        for tal_pairs in self._find_tal_pairs_for_seq():
            if tal_pairs is not None:
                tal_window_rows.extend(tal_pairs)

        df = pd.DataFrame(tal_window_rows)
        if self.outpath is not None:
            outpath = Path(self.outpath)
            outpath.parent.mkdir(parents=True, exist_ok=True)
            if df.empty:
                # Write only header
                pd.DataFrame(
                    columns=[
                        "Sequence Name",
                        "Cut Site",
                        "TAL1 start",
                        "TAL2 start",
                        "TAL1 length",
                        "TAL2 length",
                        "Spacer length",
                        "Spacer range",
                        "TAL1 RVDs",
                        "TAL2 RVDs",
                        "Plus strand sequence",
                        "Unique RE sites in spacer",
                        "% RVDs HD or NN/NH",
                    ]
                ).to_csv(outpath, sep="\t", index=False)
            else:
                df.to_csv(outpath, sep="\t", index=False)

        return df



import pytest
import tempfile
import os
import pandas as pd
from pathlib import Path
from talenWF import FindTALTask


class TestFindTALTask:
    """Test the FindTALTask class."""

    @pytest.fixture
    def sample_fasta_file(self):
        """Create a temporary FASTA file for testing with different sequences."""
        seq_name = "test_seq"
        sequence = "C" + "CT" + "C" * 26 + "T" + "C" * 26 + "CA" + "C"
        # sequence = "C" + "TC" + "C"*26 + "T" + "C"*26 + "CA"
        with tempfile.NamedTemporaryFile(mode="w", suffix=".fasta", delete=False) as f:
            f.write(f">{seq_name}\n")
            f.write(sequence + "\n")
            temp_file = f.name
        yield temp_file
        os.unlink(temp_file)

    @pytest.fixture
    def output_file(self):
        """Create a temporary output file path."""
        with tempfile.NamedTemporaryFile(suffix=".txt", delete=False) as f:
            temp_file = f.name
        yield temp_file
        if os.path.exists(temp_file):
            os.unlink(temp_file)

    def test_find_tal_task_basic(self, sample_fasta_file, output_file):
        """Test basic functionality of FindTALTask with different sequences."""
        # Create FindTALTask instance
        for min_val in [14, 18]:
            task = FindTALTask(
                fasta=sample_fasta_file,
                min_spacer=18,
                max_spacer=18,
                array_min=18,
                array_max=18,
                outpath=output_file,
                filter_base=30,
            )

            # Run the task
            task.run()

            # Check that output file was created
            assert os.path.exists(output_file)

            # Check that output file can be read properly (TSV with 2 rows skipped)
            out = pd.read_csv(output_file, delimiter="\t")
            assert (
                out.columns[0] == "Sequence Name"
            ), "First column should be Sequence Name"

            print(out[["TAL1 start", "TAL2 start"]].value_counts())

            for i, row in out.iterrows():
                print(dict(row))

            assert len(out) == 1, "Output should have only one row"
            assert (out["TAL1 start"] == 3).all(), "TAL1 start should be 3"
            assert (out["TAL2 start"] == 56).all(), "TAL2 start should be 55"

        # print(out.columns)

        # assert len(out) == 5, "Expected 5 rows"
        # for spac_len in range(14,19):
        #     assert out["Spacer length"].isin([spac_len]).any(), f"Spacer length {spac_len} not found"
        # for i,row in out.iterrows():
        #     print(row["Plus strand sequence"])
        #     assert row["Spacer length"] == len(row["Plus strand sequence"].split()[2])

    def test_find_tal_task_u_d_too_far(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "C" + "CT" + "C" * 26 + "T" + "C" * 26 + "CA" + "C"
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=17,
            array_min=14,
            array_max=18,
            outpath=output_file,
            filter_base=20,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        assert len(out) == 0, "Expected 0 rows"

    def test_find_tal_task_max_dist_odd(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "C" + "CT" + "C" * 26 + "T" + "C" * 26 + "AC" + "C"
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=17,
            array_min=14,
            array_max=18,
            outpath=output_file,
            filter_base=30,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        assert len(out) > 0, "Expected non-0 rows"
        for i, row in out.iterrows():
            print(dict(row))

    def test_find_tal_task_upstream_out_of_bounds(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "T" * 14 + "C" * 14 + "A" * 15
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=18,
            array_min=14,
            array_max=18,
            outpath=output_file,
            filter_base=22,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        assert len(out) == 0, "Expected 0 rows"

    def test_find_tal_task_upstream_at_bounds(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "T" * 15 + "C" * 14 + "A" * 15
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=18,
            array_min=14,
            array_max=18,
            outpath=output_file,
            filter_base=23,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        for i, row in out.iterrows():
            print(dict(row))
        assert len(out) == 1, "Expected exactly 1 row"

    def test_find_tal_task_upstream_out_of_bounds_nofilter(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "T" * 15 + "C" * 14 + "A" * 14
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=18,
            array_min=14,
            array_max=18,
            outpath=output_file,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        assert len(out) == 0, "Expected 0 rows"

    def test_find_tal_task_upstream_at_bounds_nofilter(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "T" * 15 + "C" * 14 + "A" * 15
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=18,
            array_min=14,
            array_max=18,
            outpath=output_file,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        print(out)
        assert len(out) == 1, "Expected 1 row"

    def test_find_tal_task_upstream_two_in_bounds_nofilter(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "T" * 16 + "C" * 14 + "A" * 15
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=18,
            array_min=14,
            array_max=18,
            outpath=output_file,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        for _, row in out.iterrows():
            print(dict(row))
        assert len(out) == 2, "Expected 1 row"

    def test_find_tal_task_upstream_in_bounds_far_nofilter(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "TT" + "C" * 54 + "A"
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=18,
            array_min=14,
            array_max=18,
            outpath=output_file,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        for _, row in out.iterrows():
            print(dict(row))
        assert len(out) == 1, "Expected 1 row"

    def test_find_tal_task_upstream_two_in_bounds_far_nofilter(self, output_file):
        """Test that the task skips TAL pairs if the upstream base is out of bounds."""
        sequence = "TT" + "C" * 53 + "A"
        task = FindTALTask(
            sequence=sequence,
            min_spacer=14,
            max_spacer=18,
            array_min=14,
            array_max=18,
            outpath=output_file,
        )
        task.run()
        assert os.path.exists(output_file)
        out = pd.read_csv(output_file, delimiter="\t")
        for _, row in out.iterrows():
            if row["Spacer range"] == "20-37":
                print(dict(row))
                # [['TAL1 start','TAL1 length','Spacer length','TAL2 length','Spacer range']]
        assert sum(out["TAL1 start"] == 1) == 1, "Expected 1 row"
        assert (
            sum((out["TAL1 start"] == 2) & (out["TAL1 length"] == 17)) > 0
        ), "Expected result(s) with TAL1 length 17"
        assert (
            sum((out["TAL1 start"] == 2) & (out["TAL2 length"] == 17)) > 0
        ), "Expected result(s) with TAL2 length 17"
        assert (
            sum((out["TAL1 start"] == 2) & (out["Spacer length"] == 17)) > 0
        ), "Expected result(s) with Spacer length 17"
        assert (
            sum(
                (out["TAL1 length"] > 18)
                | (out["TAL2 length"] > 18)
                | (out["Spacer length"] > 18)
            )
            == 0
        ), "Invalid length(s) found"


# TODO: add tests with filter base outside of the sequence
# TODO: add tests with filter base less than min_dist/2
# TODO: add tests with filter base greater than len(seq) - max_dist/2



import os
import sys
import tempfile

from talenWF.cli import main

import pandas as pd


def test_end_to_end_happy_path(monkeypatch):
    """Invoke CLI main() with all options and verify output file is created and non-empty."""
    # Create a temporary FASTA file with a simple valid sequence
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".fasta", delete=False
    ) as fasta_file:
        # Empty any preexisting contents
        fasta_file.truncate(0)
        fasta_file.write(
            ">test_seq\n" + "ATCG" * 30 + "T" + "C" * 20 + "A" + "ATCG" * 30 + "\n"
        )
        fasta_path = fasta_file.name

    # Create a temporary output file path (will be created by the CLI)
    with tempfile.NamedTemporaryFile(mode="w", suffix=".tsv", delete=False) as out_file:
        outpath = out_file.name

    try:
        # Build argv to pass all options
        argv = [
            "talenWF-findtal",
            "--fasta",
            fasta_path,
            "--min_spacer",
            "14",
            "--max_spacer",
            "18",
            "--array_min",
            "14",
            "--array_max",
            "18",
            "--outpath",
            outpath,
            "--filter_base",
            "31",
            "--upstream_bases",
            "T",
        ]

        # Patch sys.argv and invoke main
        monkeypatch.setattr(sys, "argv", argv)
        main()

        # Assert the output file is created and non-empty
        assert os.path.exists(outpath), "Output file was not created by CLI"
        assert os.path.getsize(outpath) > 0, "Output file is empty"

        out = pd.read_csv(outpath, delimiter="\t")
        assert "Sequence Name" in out.columns[0], "Column should be Sequence Name"
        assert "TAL1 start" in out.columns, "Column should be TAL1 start"
        assert "TAL2 start" in out.columns, "Column should be TAL2 start"
        assert len(out) > 0, "Output file should have at least one row"
        for i, row in out.iterrows():
            print(dict(row))

    finally:
        # Cleanup temp files
        try:
            os.unlink(fasta_path)
        except Exception:
            pass
        try:
            os.unlink(outpath)
        except Exception:
            pass


def test_end_to_end_happy_path_nofilter(monkeypatch):
    """Invoke CLI main() with all options and verify output file is created and non-empty."""
    # Create a temporary FASTA file with a simple valid sequence
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".fasta", delete=False
    ) as fasta_file:
        # Empty any preexisting contents
        fasta_file.truncate(0)
        fasta_file.write(
            ">test_seq\n" + "ATCG" * 30 + "T" + "C" * 20 + "A" + "ATCG" * 30 + "\n"
        )
        fasta_path = fasta_file.name

    # Create a temporary output file path (will be created by the CLI)
    with tempfile.NamedTemporaryFile(mode="w", suffix=".tsv", delete=False) as out_file:
        outpath = out_file.name

    try:
        # Build argv to pass all options
        argv = [
            "talenWF-findtal",
            "--fasta",
            fasta_path,
            "--min_spacer",
            "14",
            "--max_spacer",
            "18",
            "--array_min",
            "14",
            "--array_max",
            "18",
            "--outpath",
            outpath,
            "--upstream_bases",
            "T",
        ]

        # Patch sys.argv and invoke main
        monkeypatch.setattr(sys, "argv", argv)
        main()

        # Assert the output file is created and non-empty
        assert os.path.exists(outpath), "Output file was not created by CLI"
        assert os.path.getsize(outpath) > 0, "Output file is empty"

        out = pd.read_csv(outpath, delimiter="\t")
        assert "Sequence Name" in out.columns[0], "Column should be Sequence Name"
        assert "TAL1 start" in out.columns, "Column should be TAL1 start"
        assert "TAL2 start" in out.columns, "Column should be TAL2 start"
        assert len(out) > 0, "Output file should have at least one row"
        for i, row in out.iterrows():
            print(dict(row))

    finally:
        # Cleanup temp files
        try:
            os.unlink(fasta_path)
        except Exception:
            pass
        try:
            os.unlink(outpath)
        except Exception:
            pass



import argparse
import sys
import logging
from Bio import SeqIO
from .api import FindTALTask


def main():
    """CLI entry point for talenWF."""
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(levelname)s - %(message)s",
        stream=sys.stdout,
    )
    logger = logging.getLogger("talenWF")
    parser = argparse.ArgumentParser(description="talenWF: TALEN window finder")
    parser.add_argument("--fasta", required=True, help="FASTA file path")
    parser.add_argument(
        "--min_spacer", type=int, default=14, help="Minimum spacer length for TALE-NT"
    )
    parser.add_argument(
        "--max_spacer", type=int, default=18, help="Maximum spacer length for TALE-NT"
    )
    parser.add_argument(
        "--array_min", type=int, default=14, help="Minimum array length for TALE-NT"
    )
    parser.add_argument(
        "--array_max", type=int, default=18, help="Maximum array length for TALE-NT"
    )
    parser.add_argument("--outpath", default="NA", help="Output path for TALE-NT")
    parser.add_argument(
        "--filter_base",
        type=int,
        help="(1-based) Filter base position for TALE-NT (comma separated)",
    )
    parser.add_argument(
        "--upstream_bases",
        type=str,
        help="Upstream bases for TALE-NT (comma separated)",
    )
    parser.add_argument("--gspec", action="store_true", help="Use G-specific RVD")
    args = parser.parse_args()

    # Run the TAL finding task
    logger.info("Starting TAL finding task...")
    FindTALTask(
        fasta=args.fasta,
        min_spacer=args.min_spacer,
        max_spacer=args.max_spacer,
        array_min=args.array_min,
        array_max=args.array_max,
        outpath=args.outpath,
        filter_base=args.filter_base,
        upstream_bases=args.upstream_bases.split(",") if args.upstream_bases else ["T"],
        gspec=args.gspec,
    ).run()
    logger.info("TAL finding task completed successfully!")


if __name__ == "__main__":
    main()



def get_strong_binding_RVDs(gspec=False):
    strong_binding_RVDs = {"A": "NI", "C": "HD", "G": "NN", "T": "NG"}
    if gspec:
        strong_binding_RVDs["G"] = "NH"
    return strong_binding_RVDs


def get_RVD_seq(sequence, gspec=False):
    strong_binding_RVDs = get_strong_binding_RVDs(gspec)
    RVD_seq = ""
    sequence = sequence.upper()
    for i in range(len(sequence)):
        if sequence[i] not in strong_binding_RVDs:
            return None
        RVD_seq += strong_binding_RVDs[sequence[i]]
    return RVD_seq



from Bio import SeqIO
import warnings
import logging
import sys
import typing

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s - %(message)s",
    stream=sys.stdout,
)
logger = logging.getLogger("talenWF")


def get_sequence(
    fasta: typing.Optional[str] = None, sequence: typing.Optional[str] = None
) -> tuple[str, str]:
    """
    Get the sequence from a FASTA file or a string.

    Args:
        fasta: The path to the FASTA file.
        sequence: The sequence string.
    Returns:
        tuple[str, str]: The sequence and sequence ID.
    """
    sequence_id = None
    if fasta is not None:
        # Validate FASTA file exists and is readable
        try:
            with open(fasta, "r") as f:
                # Try to parse the first record to validate FASTA format
                records_iter = SeqIO.parse(f, "fasta")
                first_record = next(records_iter)
                logger.info(f"Processing FASTA file: {fasta}")
                logger.info(
                    f"First sequence: {first_record.id} (length: {len(first_record.seq)})"
                )
                sequence = str(first_record.seq).upper()
                sequence_id = first_record.id
                if next(records_iter, None) is not None:
                    warnings.warn(
                        "Multiple sequences found in FASTA file, only the first sequence will be used"
                    )
        except FileNotFoundError:
            raise FileNotFoundError(f"FASTA file not found: {fasta}")
        except Exception as e:
            raise Exception(f"Error reading FASTA file: {e}")
    else:
        # use sequence arg if fasta not provided
        if sequence is None:
            raise ValueError("Sequence is required if fasta is not provided")
        sequence = sequence.upper()
        sequence_id = "Sequence"

    if len(sequence) == 0:
        raise ValueError("Sequence is empty")
    return sequence, sequence_id
